%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   Matlab Codes associated to the paper
%  Langevin and Hamiltonian based Sequential MCMC 
%   for Efficient Bayesian filtering in High-Dimensional Spaces
%
%       
%       ArXiv
%   
%       Author: F. Septier & G. W. Peters
%       Date: 09/03/2015
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear all
close all

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Definitions of the parameters of the scenarios
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
NbMcRuns=30;


ModelParam.Dimension=8^2; % Dimension of X_k
% By increasing the size of the domain with dimension I keep the
% correlation between 2 consecutives cases equal
ModelParam.Domain=[1 sqrt(ModelParam.Dimension)];%[0 30];

ModelParam.NbSensors=ModelParam.Dimension; % Dimension of Y_k 
ModelParam.NoTimeSteps=10; % No Of Time Steps


% Sensors placed on a regular grid
[Xsensors,YSensors]=meshgrid(linspace(ModelParam.Domain(1),ModelParam.Domain(2),sqrt(ModelParam.NbSensors)));
ModelParam.SensorPosition=reshape(Xsensors,1,[]);
ModelParam.SensorPosition=[ModelParam.SensorPosition;reshape(YSensors,1,[])];


ModelParam.TransitionMatrix=0.9*eye(ModelParam.Dimension);
ModelParam.ObservationTransition=eye(ModelParam.Dimension); % Required to Map supp(X_k) to supp(Y_k)


ModelParam.StateCovParam(1)=3; % Variance
ModelParam.InitialMean=log(10)*ones(ModelParam.Dimension,1);%log(100)*ones(ModelParam.Dimension,1);
ModelParam.StateCovParam(2)=20; % Length scale of the exp in Correlation
ModelParam.StateCovParam(3)=0.01; %in Normal/Normal
ModelParam.StateCovariance=KernelStateDynamics(ModelParam.SensorPosition,ModelParam.StateCovParam);


ModelParam.StateCovariance=ModelParam.StateCovariance+ModelParam.StateCovParam(3)*eye(ModelParam.Dimension);


ModelParam.StateCovariance=triu(ModelParam.StateCovariance,1)+triu(ModelParam.StateCovariance,1)'+diag(diag(ModelParam.StateCovariance));
ModelParam.StateCovarianceSR=real(sqrtm(ModelParam.StateCovariance));
ModelParam.StateCovarianceInv=real(inv(ModelParam.StateCovariance));
ModelParam.StateCovarianceDet=real(det(ModelParam.StateCovariance));

ModelParam.StateSkewness=0.3*ones(ModelParam.Dimension,1); % Skewness when GH Skewed-t Used
ModelParam.StateDegreeFreedom=7;% \nu Degree Freedom when GH Skewed-t Used
% Computation of Covariance SkewedtT
a=ModelParam.StateDegreeFreedom/2;
b=a;
MeanW=b/(a-1);
VarW=(b^2)/(((a-1)^2)*(a-2));
ModelParam.StateCovSkewedT=MeanW*ModelParam.StateCovariance+VarW*ModelParam.StateSkewness*ModelParam.StateSkewness';
ModelParam.InvStateCovSkewedT=inv(ModelParam.StateCovSkewedT);   


ModelParam.MappingIntensityFunction=@(x,Coeff,Scale) Coeff*exp(x/Scale);
ModelParam.MappingIntensityFunctionFirstDeriv=@(x,Coeff,Scale) Coeff*exp(x/Scale)/Scale;
ModelParam.MappingIntensityFunctionSecDeriv=@(x,Coeff,Scale) Coeff*exp(x/Scale)/(Scale^2);
ModelParam.MappingIntensityFunctionThirdDeriv=@(x,Coeff,Scale) Coeff*exp(x/Scale)/(Scale^3);
ModelParam.MappingIntensityFunctionLog=@(x,Coeff,Scale) log(Coeff)+(x/Scale);
ModelParam.MappingIntensityFunctionFirstDerivLog=@(x,Coeff,Scale) (1/Scale)*ones(size(x));
ModelParam.MappingIntensityFunctionSecDerivLog=@(x,Coeff,Scale) zeros(size(x));



 
ModelParam.MappingIntensityCoeff=1;
ModelParam.MappingIntensityScale=3;

ModelParam.VarNoise=2; % Variance in the Normal Observation Case


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Choice between different likelihood
% 'Poisson' -> Indepedent Poisson Observations
% 'Normal' -> Indepedent Normal Observations
%
% Choice between different State
% 'SkewedT' -> Multivariate Skewed T distribution
% 'Normal' -> Multivariate Normal Observations
ModelParam.ModelObservation='Normal';%'Poisson'
ModelParam.ModelState='Normal';%'SkewedT'

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Folder in which the results are saved
FolderNameSavedData=['SavedData/'];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Folder in which the Data set generated are saved
FolderNameSavedSyntheticData=['SavedGeneratedData/'];

StudyVarianceEstimator=0; % If 1 the algorithms will be repeated on the same data set

for NumMCruns=1:NbMcRuns
    % Example Multivariate Draw from that


    State=zeros(ModelParam.Dimension,ModelParam.NoTimeSteps);
    
    
    % Test if Data has been already generated for this scenario
    NameDataPrior=['_Var=',num2str(ModelParam.StateCovParam(1)),'_Corr=',num2str(ModelParam.StateCovParam(2))];
    if strcmp(ModelParam.ModelState,'SkewedT')
        NameDataPrior=[NameDataPrior,'_Skew=',num2str(ModelParam.StateSkewness(1)),'_DegFree=',num2str(ModelParam.StateDegreeFreedom)];
    end
    if strcmp(ModelParam.ModelObservation,'Poisson')
        NameDataObs=['_IntCoeff=',num2str(ModelParam.MappingIntensityCoeff),'_Scale=',num2str(ModelParam.MappingIntensityScale)];
    elseif strcmp(ModelParam.ModelObservation,'Normal')
        NameDataObs=['_Var=',num2str(ModelParam.VarNoise)];
    end
    FileNameToTest=['Data_NbTime=',num2str(ModelParam.NoTimeSteps),'_Dim=',num2str(ModelParam.Dimension),'_Prior=',ModelParam.ModelState,NameDataPrior,'_Obs=',ModelParam.ModelObservation,NameDataObs,'_NumMCruns='];
    
    
    if StudyVarianceEstimator==1
        NumDataSet=1;
    else
        NumDataSet=NumMCruns;
    end
    
    if exist([FolderNameSavedSyntheticData,FileNameToTest,num2str(NumDataSet),'.mat'])~=0
        disp('Data Already Exist')
        eval(['load ',FolderNameSavedSyntheticData,FileNameToTest,num2str(NumDataSet),'.mat'])
             
    else % Generation of the data set
        for t=1:ModelParam.NoTimeSteps
            

            if t==1
                if strcmp(ModelParam.ModelState,'SkewedT')
                    W=1./gamrnd(ModelParam.StateDegreeFreedom/2,1/(ModelParam.StateDegreeFreedom/2));
                    State(:,t)=ModelParam.InitialMean+W*ModelParam.StateSkewness+sqrt(W)*ModelParam.StateCovarianceSR*randn(ModelParam.Dimension,1);
                elseif strcmp(ModelParam.ModelState,'Normal')
                    State(:,t)=ModelParam.InitialMean+ModelParam.StateCovarianceSR*randn(ModelParam.Dimension,1);
                end
            else
                if strcmp(ModelParam.ModelState,'SkewedT')
                    W=1./gamrnd(ModelParam.StateDegreeFreedom/2,1/(ModelParam.StateDegreeFreedom/2));
                    State(:,t)=ModelParam.TransitionMatrix*State(:,t-1)+W*ModelParam.StateSkewness+sqrt(W)*ModelParam.StateCovarianceSR*randn(ModelParam.Dimension,1);
                elseif strcmp(ModelParam.ModelState,'Normal')
                    State(:,t)=ModelParam.TransitionMatrix*State(:,t-1)+ModelParam.StateCovarianceSR*randn(ModelParam.Dimension,1);
                end
                
            end

            if strcmp(ModelParam.ModelObservation,'Poisson')
                
                IntensityPoisson=ModelParam.MappingIntensityFunction(ModelParam.ObservationTransition*State(:,t),ModelParam.MappingIntensityCoeff,ModelParam.MappingIntensityScale);
                Observation(:,t)=poissrnd(IntensityPoisson);
                
            elseif strcmp(ModelParam.ModelObservation,'Normal')

                % Normal Observations

                Observation(:,t)=ModelParam.ObservationTransition*State(:,t)+sqrt(ModelParam.VarNoise)*randn(ModelParam.Dimension,1);


            end


            % To Display the generated Data Set
            if 0
                figure(1);plot(ModelParam.MappingIntensityFunction(State,ModelParam.MappingIntensityCoeff,ModelParam.MappingIntensityScale)');title('Intensity')
                figure(2);plot(IntensityPoisson)
                figure(3);imagesc(ModelParam.SensorPosition(1,:),ModelParam.SensorPosition(2,:),reshape(State(:,t),sqrt(ModelParam.Dimension),[]));
                %colormap gray;
                axis xy;

                 figure(4);imagesc(ModelParam.SensorPosition(1,:),ModelParam.SensorPosition(2,:),reshape(State(:,t),sqrt(ModelParam.Dimension),[]));
                axis xy;

                figure(6);surfc(Xsensors,YSensors,reshape(State(:,t),sqrt(ModelParam.Dimension),[]));%axis xy;
                figure(5);imagesc(ModelParam.SensorPosition(1,:),ModelParam.SensorPosition(2,:),reshape(Observation(:,t),sqrt(ModelParam.Dimension),[]));axis xy;
                title('Observed Count Process')
                pause(0.1)
            end
            
        end % End Loop over Time step
        
        eval(['save ',FolderNameSavedSyntheticData,FileNameToTest,num2str(NumDataSet),'.mat ModelParam Observation State'])
    end
    

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %    Kalman - In case of Normal/Normal Model
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
    if strcmp(ModelParam.ModelObservation,'Normal') && strcmp(ModelParam.ModelState,'Normal')
        tic
        [Kalman]=KalmanFilter(ModelParam,Observation);
        AlgoAnalysis.Kalman.AllData=Kalman;
        AlgoAnalysis.Kalman.Time(NumMCruns)=toc;
        AlgoAnalysis.Kalman.MSE(NumMCruns)=mean(mean((State-Kalman.UpdateMean).^2));        
    end
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %    SMCMC
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Definition of the parameters of the SMCMC filter
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    BlockSize=ModelParam.Dimension; % Size Block in the refinement of SMCMC
    mParticles=200; % Number of particles that will be used in all Monte-Calro algorithms
    AlgoSeqMCMC.NoParticles=mParticles; 
    AlgoSeqMCMC.NoBurnIn=mParticles/10; % BurnIn Period SMCMC Set to 10% of NoParticles
    AlgoSeqMCMC.Thining=1;

    AlgoSeqMCMC.BothJointDraw_RefinementAtEachIteration=1; % 1 - Do both JD & Refin. Otherwise do JD with ProbaJointDraw and Ref. with Proba 1-ProbaJointDraw
    AlgoSeqMCMC.ProbaJointDraw=0.75;
    AlgoSeqMCMC.BlockSize=BlockSize; % Size of the Block in the refinement Step
    
    
    ModelParam.AlphaPosDefMatrix=1e5;% To ensure positive definite matrix for the negative hessian of the log prior !!
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Choice between different techniques
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 'PRIOR' -> Use independent proposal based on cond. prior
    % 'optimal' -> only for Normal/Normal Case - Equivalent to Fully adapted Particle filter
    % 'SMALA' -> Seq.  MALA with a Pre conditionned matrix, i.e. Unitary Identity Matrix
    % 'SimplSmMALA' -> Seq. Manifold MALA with no drift 
    % 'SmMALA' -> Seq. Manifold MALA 
    % 'SHMC' -> HMC sampler with preconditionned Matrix to be defined below
    % 'SmHMC' -> Seq. Manifold HMC
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    AlgoSeqMCMC.SamplingTechnique='SmHMC'; % Choose of the MCMC Kernel
    
    
    AlgoSeqMCMC.NoLeapfrogSteps=20;% Number of Leapfrog Step in HMC-based sampling
    AlgoSeqMCMC.StepSize=0.5;
    
    
    
    AlgoSeqMCMC.NoFixedPointGeneLeapfrog=2; % Number of fixed point for Generalized Leapfrog
    AlgoSeqMCMC.DisplayProposalCandidate=0; % To display at each MCMC iteration what are the proposed value for x

    % Use in HMC - Preconditionned Matrix for Momentum Variables
    AlgoSeqMCMC.CovAuxVarHMC=(1)*eye(ModelParam.Dimension);
    AlgoSeqMCMC.InvCovAuxVarHMC=inv(AlgoSeqMCMC.CovAuxVarHMC);
    AlgoSeqMCMC.SRInvCovAuxVarHMC=sqrtm(AlgoSeqMCMC.CovAuxVarHMC);
    
    % To indicate wether the generalized leapfrog is necessary when mHMC is employed 
    % Indeed not required when the metric does not depend on the state !!! Accelerate if not used 
    % Example Normal/Normal Model || Normal/Poisson if the metric is
    % obtained by expectation over the obs and data !
    ModelParam.PoissonNormalPositionDependent=1; % if 1 we take the FIM of the poisson without marginalizing the X otherwise we marginalize with respect to X;
    if (strcmp(ModelParam.ModelObservation,'Poisson') ) || strcmp(ModelParam.ModelState,'SkewedT')
        AlgoSeqMCMC.GeneralizedLeapfrogRequired=1;
    elseif strcmp(ModelParam.ModelObservation,'Poisson') && strcmp(ModelParam.ModelState,'Normal') && ModelParam.PoissonNormalPositionDependent==1
        AlgoSeqMCMC.GeneralizedLeapfrogRequired=1;
    else
        AlgoSeqMCMC.GeneralizedLeapfrogRequired=0;
    end
    
    
    
    tic;
    [SeqMCMCParticles]=SequentialMCMC_JointPosteriorv1(ModelParam,Observation,AlgoSeqMCMC,State);
    AlgoAnalysis.SeqMCMC.AllData=SeqMCMCParticles;
    AlgoAnalysis.SeqMCMC.Time(NumMCruns)=toc;
    AlgoAnalysis.SeqMCMC.MSE(NumMCruns)=mean(mean((State-SeqMCMCParticles.Mean).^2));
    AlgoAnalysis.SeqMCMC.MSE2(NumMCruns)=mean(sum((State-SeqMCMCParticles.Mean).^2,1));
    AlgoAnalysis.SeqMCMC.AcceptanceRate(NumMCruns)=SeqMCMCParticles.Analysis.AcceptationRefinement/SeqMCMCParticles.Analysis.NoRefinement;
    

    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % --  SIR Algorithm ---
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    AlgoSIR=AlgoSeqMCMC;
    AlgoSIR.ResamplingEachTimeStep=0;
    AlgoSIR.ResamplingThreshold=AlgoSIR.NoParticles/2;
    
    tic;
    [Particles]=ParticleFilter(ModelParam,Observation,AlgoSIR,State);
    AlgoAnalysis.SIR.AllData=Particles;
    AlgoAnalysis.SIR.Time(NumMCruns)=toc;
    AlgoAnalysis.SIR.MSE(NumMCruns)=mean(mean((State-Particles.Mean).^2));
 
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % -- BLOCK SIR Resample MOVE (Resampling at each Time)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    AlgoBlockSIR=AlgoSIR;% Same parameters as the SIR 

    AlgoBlockSIR.BlockSize=2^2;% Size of the block
    AlgoBlockSIR.NoBlocks=ModelParam.Dimension/AlgoBlockSIR.BlockSize;    

    % To have blocks that are spatially coherent...
    AlgoBlockSIR.NumeroBlock=reshape(kron(reshape(1:AlgoBlockSIR.NoBlocks,sqrt(AlgoBlockSIR.NoBlocks),sqrt(AlgoBlockSIR.NoBlocks)),ones(sqrt(AlgoBlockSIR.BlockSize),sqrt(AlgoBlockSIR.BlockSize))),[],1);
    [B,I]=sort(AlgoBlockSIR.NumeroBlock);
    AlgoBlockSIR.IndexForBlocks=I';
    



    tic;
    [BlockSIRParticles]=BlockParticleFilter(ModelParam,Observation,AlgoBlockSIR,State);            
    AlgoAnalysis.BlockSIR.AllData=BlockSIRParticles;
    AlgoAnalysis.BlockSIR.Time(NumMCruns)=toc;
    
    AlgoAnalysis.BlockSIR.MSE(NumMCruns)=mean(mean((State-BlockSIRParticles.Mean).^2));
    
    disp('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
    disp('%%%%%------- Results --------%%%%%%')
    disp('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
    disp(['%%%%%------- Run ',num2str(NumMCruns),'/',num2str(NbMcRuns),' --------%%%%%%'])
    if strcmp(ModelParam.ModelObservation,'Normal') && strcmp(ModelParam.ModelState,'Normal')
        
        disp('Kalman Filter')
        disp(['Average Computation Time: ',num2str(mean(AlgoAnalysis.Kalman.Time(1:NumMCruns)))])
        disp(['Average MSE: ',num2str(mean(AlgoAnalysis.Kalman.MSE(1:NumMCruns)))])
        disp('-------------')        
        
    end
    disp('Particle Filter')
    disp(['Average Computation Time: ',num2str(mean(AlgoAnalysis.SIR.Time(1:NumMCruns)))])
    disp(['Average MSE: ',num2str(mean(AlgoAnalysis.SIR.MSE(1:NumMCruns)))])
    disp('-------------')
    disp('Block Particle Filter')
    disp(['Average Computation Time: ',num2str(mean(AlgoAnalysis.BlockSIR.Time(1:NumMCruns)))])
    disp(['Average MSE: ',num2str(mean(AlgoAnalysis.BlockSIR.MSE(1:NumMCruns)))])
    disp('-------------')
    disp(['Sequential MCMC with ',AlgoSeqMCMC.SamplingTechnique])
    disp(['Average Computation Time: ',num2str(mean(AlgoAnalysis.SeqMCMC.Time(1:NumMCruns)))])
    disp(['Average MSE: ',num2str(mean(AlgoAnalysis.SeqMCMC.MSE(1:NumMCruns)))])
    disp(['Average Accept. Rate: ',num2str(mean(AlgoAnalysis.SeqMCMC.AcceptanceRate(1:NumMCruns)))])
    

    
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Save Results at Each MC iterations
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    if AlgoSeqMCMC.GeneralizedLeapfrogRequired==1
        StringToAdd='_GeneralizedLF';
    else
        StringToAdd='_LF';
    end
    if StudyVarianceEstimator==1
        StingToSave=['Results_StudyVarWithNumDataSet=',num2str(NumDataSet),'_'];
    else
        StingToSave=['Results_'];
    end
    
    StingToSave=[StingToSave,FileNameToTest(1:end-10),'_AlgoSMCMCwith',AlgoSeqMCMC.SamplingTechnique,...    
        StringToAdd,'_NbPart=',num2str(AlgoSeqMCMC.NoParticles),'_NumMCruns=',num2str(NumMCruns),'.mat'];

    eval(['save ',FolderNameSavedData,StingToSave])
    
end


